from ..base import rules


class AntBannedGroup(rules.BannedGroup):
    banned_groups = {
        '3LTON',
        '4yEo',
        'ADE',
        'AFG',
        'AniHLS',
        'AnimeRG',
        'AniURL',
        'AROMA',
        'aXXo',
        'Brrip',
        'CHD',
        'CM8',
        'CrEwSaDe',
        'd3g',
        'DDR',
        'DeadFish',
        'DNL',
        'ELiTE',
        'eSc',
        'FaNGDiNG0',
        'FGT',
        'Flights',
        'FRDS',
        'FUM',
        'HAiKU',
        'HD2DVD',
        'HDS',
        'HDTime',
        'Hi10',
        'ION10',
        'iPlanet',
        'JIVE',
        'KiNGDOM',
        'Leffe',
        'LiGaS',
        'LOAD',
        'MeGusta',
        'mHD',
        'MkvCage',
        'mSD',
        'NhaNc3',
        'nHD',
        'NOIVTC',
        'nSD',
        'Oj',
        'Ozlem',
        'PiRaTeS',
        'PRoDJi',
        'RAPiDCOWS',
        'RARBG',
        'RDN',
        'REsuRRecTioN',
        'RetroPeeps',
        'RMTeam',
        'SANTi',
        'SicFoI',
        'SPASM',
        'SPDVD',
        'STUTTERSHIT',
        'TBS',
        'Telly',
        'TM',
        'UPiNSMOKE',
        'URANiME',
        'WAF',
        'xRed',
        'XS',
        'YIFY',
        'YTS',
        'Zeus',
        'ZKBL',
        'ZmN',
        'ZMNT',
    }
